# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/losses.ipynb.

# %% ../nbs/losses.ipynb 2
from __future__ import annotations

import torch.nn.modules.loss as TL

from .imports import *

# %% auto 0
__all__ = ['BCEWithLogitsLoss', 'ClassBalancedCrossEntropyLoss', 'ClassBalancedBCEWithLogitsLoss']

# %% ../nbs/losses.ipynb 4
class BCEWithLogitsLoss(TL._Loss):
    """
    Like `nn.BCEWithLogitsLoss`, but with 'batchmean' reduction from MosiacML. `batchmean` scales
    loss by the batch size which results in larger loss values more similar to `nn.CrossEntropy`
    then `mean` reduction.
    """
    def __init__(self,
        weight:Tensor|None=None, # Rescaling weieght for each class
        reduction:str='mean', # Reduction to apply to loss output. Also supports 'batchmean'.
        pos_weight:Tensor|None=None # Weight of positive examples
    ) -> None:
        super().__init__(None, None, reduction)
        self.register_buffer('weight', weight)
        self.register_buffer('pos_weight', pos_weight)
        self.weight:Tensor|None
        self.pos_weight:Tensor|None

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if self.reduction == 'batchmean':
            return F.binary_cross_entropy_with_logits(input, target,
                                                      self.weight,
                                                      pos_weight=self.pos_weight,
                                                      reduction='sum')/torch.tensor(input.shape[0])
        else:
            return F.binary_cross_entropy_with_logits(input, target,
                                                      self.weight,
                                                      pos_weight=self.pos_weight,
                                                      reduction=self.reduction)

    def decodes(self, x:Tensor) -> Tensor:
        "Converts model output to target format"
        return x>self.thresh

    def activation(self, x:Tensor) -> Tensor:
        "`nn.BCEWithLogitsLoss`'s fused activation function applied to model output"
        return torch.sigmoid(x)

# %% ../nbs/losses.ipynb 6
# Refactored from https://github.com/vandit15/Class-balanced-loss-pytorch
# Class-balanced-loss-pytorch - MIT License - Copyright (c) 2019 Vandit Jain

class ClassBalanced(TL._WeightedLoss):
    "Class Balanced weight calculation, from https://arxiv.org/abs/1901.05555."
    def __init__(self,
        samples_per_class:Tensor,
        beta:float=0.99,
        reduction:str='mean',
    ):
        num_classes = len(samples_per_class)
        if not isinstance(samples_per_class, Tensor):
            samples_per_class = Tensor(samples_per_class)

        effective_num = 1.0 - torch.pow(beta, samples_per_class)
        weight = (1.0 - beta) / effective_num
        weight = weight / weight.sum() * num_classes

        super().__init__(weight, None, None, reduction)

    @delegates(nn.Module.to)
    def to(self, *args, **kwargs):
        device, dtype, non_blocking, convert_to_format = torch._C._nn._parse_to(*args, **kwargs)
        if self.weight.device != device: self.weight = self.weight.to(device=device)
        super().to(*args, **kwargs)

# %% ../nbs/losses.ipynb 7
class ClassBalancedCrossEntropyLoss(ClassBalanced):
    "Class Balanced Cross Entropy Loss, from https://arxiv.org/abs/1901.05555."
    __constants__ = ['ignore_index', 'reduction', 'label_smoothing']
    ignore_index: int
    label_smoothing: float

    def __init__(self,
        samples_per_class:Tensor,
        beta:float=0.99,
        ignore_index:int=-100,
        reduction:str='mean',
        label_smoothing:float=0.0,
        axis:int=-1
    ):
        super().__init__(samples_per_class, beta, reduction)
        self.ignore_index = ignore_index
        self.label_smoothing = label_smoothing
        self.axis=axis

    def forward(self, input:Tensor, target:Tensor) -> Tensor:
        return F.cross_entropy(input, target, weight=self.weight,
                               ignore_index=self.ignore_index, reduction=self.reduction,
                               label_smoothing=self.label_smoothing)

    def decodes(self, x:Tensor):
        "Converts model output to target format"
        return x.argmax(dim=self.axis)

    def activation(self, x:Tensor):
        "`nn.CrossEntropyLoss`'s fused activation function applied to model output"
        return F.softmax(x, dim=self.axis)

# %% ../nbs/losses.ipynb 8
class ClassBalancedBCEWithLogitsLoss(ClassBalanced):
    "Class Balanced BCE With Logits Loss, from https://arxiv.org/abs/1901.05555 with 'batchmean' reduction"

    def __init__(self,
        samples_per_class:Tensor,
        beta:float=0.99,
        reduction:str='mean',
        thresh:float=0.5, # Threshold for `decodes`
    ):
        super().__init__(samples_per_class, beta, reduction)
        self.thresh=thresh

    def forward(self, input:Tensor, target:Tensor) -> Tensor:
        if self.reduction == 'batchmean':
            return F.binary_cross_entropy_with_logits(input, target,
                                                      self.weight,
                                                      pos_weight=self.pos_weight,
                                                      reduction='sum')/torch.tensor(input.shape[0])
        else:
            return F.binary_cross_entropy_with_logits(input, target,
                                                      self.weight,
                                                      pos_weight=self.pos_weight,
                                                      reduction=self.reduction)

    def decodes(self, x:Tensor) -> Tensor:
        "Converts model output to target format"
        return x>self.thresh

    def activation(self, x:Tensor) -> Tensor:
        "`nn.BCEWithLogitsLoss`'s fused activation function applied to model output"
        return torch.sigmoid(x)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/schedulers.ipynb.

# %% ../nbs/schedulers.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../nbs/schedulers.ipynb 3
from __future__ import annotations

import numpy as np

from fastcore.basics import even_mults

from fastai.callback.core import Callback
from fastai.callback.schedule import SchedCos, SchedNo, combine_scheds, ParamScheduler, _Annealer
from fastai.learner import Learner

from .basics import is_listish
from .imports import *

# %% auto 0
__all__ = []

# %% ../nbs/schedulers.ipynb 6
@patch
def fit_flat_warmup(self:Learner,
    n_epoch:int, # Number of epochs
    lr:float|None=None, # Maximum learning rate
    div:Numeric=25., # Initial learning rate: `lr/div`
    div_final:Numeric=1e5, # Final learning rate: `lr/div_final`
    pct_start:float=0.75, # Start learning rate cosine annealing
    warm_pct:float=0.2, # Learning rate warmup in percent
    warm_epoch:int=5, # Learning rate warmup in epochs
    warm_mode:str='auto', # Warmup using 'epoch', 'pct', or min of epoch/pct if 'auto'
    warm_sched:Callable[..., _Annealer]=SchedCos, # Learning rate warmup schedule
    wd:float|None=None, # Weight decay, defaults to `Optimizer` weight decay
    cbs:Listified[Callback]|None=None, # Temporary Callbacks to apply during fit
    reset_opt:bool=False # Reset `Optimizer` before fit
):
    "Fit `self.model` for `n_epoch` at flat `lr` with a warmup and ending with cosine annealing."
    if self.opt is None:
        self.create_opt()
    self.opt.set_hyper('lr', self.lr if lr is None else lr)

    if warm_mode=='auto':
        warm_pct = min(warm_epoch/n_epoch, warm_pct)
    elif warm_mode=='epoch':
        warm_pct = warm_epoch/n_epoch

    pcts = [warm_pct, 1-(warm_pct+(1-pct_start)), 1-pct_start]
    scheds = [warm_sched(lr/div, lr), SchedNo(lr,lr), SchedCos(lr, lr/div_final)]
    scheds = {'lr': combine_scheds(pcts, scheds)}

    self.fit(n_epoch, cbs=ParamScheduler(scheds)+L(cbs), reset_opt=reset_opt, wd=wd, start_epoch=0)

# %% ../nbs/schedulers.ipynb 11
@patch
def fit_cos_anneal(self:Learner,
    n_epoch:int, # Number of epochs
    lr:float|None=None, # Maximum learning rate
    div:Numeric=25., # Initial learning rate: `lr/div`
    div_final:Numeric=1e5, # Final learning rate: `lr/div_final`
    warm_pct:float=0.2, # Learning rate warmup in percent
    warm_epoch:int=5, # Learning rate warmup in epochs
    warm_mode:str='auto', # Warmup using 'epoch', 'pct', or min of epoch/pct 'auto'
    warm_sched:Callable[..., _Annealer]=SchedCos, # Learning rate warmup schedule
    wd:float|None=None, # Weight decay, defaults to `Optimizer` weight decay
    cbs:Listified[Callback]|None=None, # Temporary Callbacks to apply during fit
    reset_opt:bool=False # Reset `Optimizer` before fit
):
    "Fit `self.model` for `n_epoch` using a with cosine annealing schedule with a max `lr` and optional warmup."
    if self.opt is None:
        self.create_opt()
    self.opt.set_hyper('lr', self.lr if lr is None else lr)

    if warm_mode=='auto':
        warm_pct = min(warm_epoch/n_epoch, warm_pct)
    elif warm_mode=='epoch':
        warm_pct = warm_epoch/n_epoch

    if warm_pct > 0:
        pcts = [warm_pct, 1-warm_pct]
        scheds = [warm_sched(lr/div, lr), SchedCos(lr, lr/div_final)]
    else:
        pcts = [1]
        scheds = [SchedCos(lr, lr/div_final)]
    scheds = {'lr': combine_scheds(pcts, scheds)}

    self.fit(n_epoch, cbs=ParamScheduler(scheds)+L(cbs), reset_opt=reset_opt, wd=wd, start_epoch=0)

# %% ../nbs/schedulers.ipynb 19
@patch
def fit_flat_varied(self:Learner,
    n_epoch:int, # Number of epochs
    start_lr:float|None=None, # Initial learning rate
    div_final:Numeric=1e5, # Final learning rate: `lr/div_final`
    pct_start:float=0.75, # Start learning rate cosine annealing
    wd:float|None=None, # Weight decay, defaults to `Optimizer` weight decay
    next_lr:Listified[float]|slice|None=None, # Learning rates to switch to at `change_by`. Must be same length as `change_by`
    change_by:Listified[int]|Listified[float]|None=None, # Epochs or percent of steps to switch to `next_lr` by. Must be same length as `next_lr`
    change_time:Listified[int]|Listified[float]=1, # If greater than 0 (percent of steps or epochs), how long to change to `next_lr`. Must be same length as `next_lr`
    change_sched:Listified[Callable[..., _Annealer]]|None=None, # Schedule(s) for change. Defaults to `SchedCos`. Must be same length as `next_lr`
    cbs:Listified[Callback]|None=None, # Temporary Callbacks to apply during fit
    reset_opt:bool=False # Reset `Optimizer` before fit
):
    """
    Fit `self.model` for `n_epoch` at flat `start_lr`, then change to flat `next_lr` at `change_by`,
    optionally with cosine annealing or custom `change_sched` over `change_time`. Final cosine annealing at `pct_start`.
    """
    assert isinstance(next_lr, (float, slice)) or (is_listish(next_lr) and len(next_lr)>=1), f'{next_lr=} must be float, slice, or list of float or slice'
    assert isinstance(change_by, (int, float, slice)) or (is_listish(change_by) and len(change_by)>=1), f'{change_by=} must be int, float, slice, or list of int, float, or slice'

    if self.opt is None:
        self.create_opt()
    self.opt.set_hyper('lr', self.lr if start_lr is None else start_lr)
    start_lr = np.array([h['lr'] for h in self.opt.hypers])
    params_len = len(start_lr)

    if not is_listish(next_lr):
        next_lr = [next_lr]
    if not is_listish(change_by):
        change_by = [change_by]
    change_by = [i/n_epoch if i>=1 else i for i in change_by]
    assert len(change_by)==len(next_lr), f'{next_lr=} & {change_by=} need to be same length'

    if not is_listish(change_time):
        change_time = [change_time]*len(change_by)
    else: assert len(change_by)==len(change_time), f'{change_time=} list needs to be same length as {next_lr=} & {change_by=}'
    change_time = [i/n_epoch if i>=1 else i for i in change_time]

    if change_sched is not None:
        if not is_listish(change_sched):
            change_sched = [change_sched]
        assert len(change_by)==len(change_sched), f'{next_lr=} & {change_sched=} need to be same length'

    pcts, scheds, last_lr, last_pct = [], [SchedNo(start_lr, start_lr)], start_lr, 0
    for i, cb in enumerate(change_by):
        if cb < pct_start:
            nlr = next_lr[i]
            if isinstance(nlr, slice):
                if nlr.start: nlr = even_mults(nlr.start, nlr.stop, params_len)
                else: nlr = [nlr.stop/10]*(params_len-1) + [nlr.stop]
                nlr=np.array(nlr)

            change_pct = cb - change_time[i]
            assert change_pct >= last_pct, f'{change_pct=} in pos {i} of {change_by=} overlaps with previous schedule {last_pct}'

            pcts.append(change_pct - sum(pcts))
            scheds.append(SchedNo(nlr, nlr))

            if change_time[i] > 0:
                pcts.append(change_time[i])
                if is_listish(change_sched): scheds.insert(-1, change_sched[i](last_lr, nlr))
                else: scheds.insert(-1, SchedCos(last_lr, nlr))

            last_lr = nlr
            last_pct = change_pct
        else:
            warn(f'change_by: {change_by[i]} is after {pct_start=} and ignored.')

    pcts += [pct_start - sum(pcts), 1-pct_start]
    scheds += [SchedCos(last_lr, last_lr/div_final)]
    scheds = {'lr': combine_scheds(pcts, scheds)}

    self.fit(n_epoch, cbs=ParamScheduler(scheds)+L(cbs), reset_opt=reset_opt, wd=wd)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/transform.ipynb.

# %% ../nbs/transform.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../nbs/transform.ipynb 3
from __future__ import annotations
from torch.distributions import Bernoulli

from fastcore.transform import DisplayedTransform, _is_tuple, retain_type
from fastcore.dispatch import typedispatch, explode_types

from fastai.data.core import TfmdDL

from .imports import *

# %% auto 0
__all__ = ['BatchRandTransform']

# %% ../nbs/transform.ipynb 5
@patch
def to(self:TfmdDL, device):
    self.device = device
    for tfm in self.after_batch.fs:
        for a in L(getattr(tfm, 'parameters', None)): setattr(tfm, a, getattr(tfm, a).to(device))
        if hasattr(tfm, 'to'): tfm.to(device)
    return self

# %% ../nbs/transform.ipynb 6
@patch
def _one_pass(self:TfmdDL):
    b = self.do_batch([self.do_item(None)])
    if self.device is not None:
        b = to_device(b, self.device)
        self.to(self.device)
    its = self.after_batch(b)
    self._n_inp = 1 if not isinstance(its, (list,tuple)) or len(its)==1 else len(its)-1
    self._types = explode_types(its)

# %% ../nbs/transform.ipynb 8
class BatchRandTransform(DisplayedTransform):
    "Randomly selects a subset of batch `b` to apply transform with per item probability `p` in `before_call`"
    do,supports,split_idx = True,[],0
    def __init__(self,
        p:Numeric=1., # Probability of applying Transform to each batch item
        before_call:Callable[[Tensor|tuple[Tensor,...],int],None]|None=None, # Batch preprocessing function
        **kwargs
    ):
        store_attr('p')
        super().__init__(**kwargs)
        self.before_call = ifnone(before_call,self.before_call)
        self.bernoulli = Bernoulli(p)

    def before_call(self,
        b:Tensor|tuple[Tensor,...], # Batch item(s)
        split_idx:int # Train (0) or valid (1) index
    ):
        "Randomly select `self.idxs` and set `self.do` based on `self.p` if not valid `split_idx`"
        self.idxs = self.bernoulli.sample((find_bs(b),)).bool() if not split_idx and self.p<1. else torch.ones(find_bs(b)).bool()
        self.do = self.p==1. or self.idxs.shape[-1] > 0

    def __call__(self,
        b:Tensor|tuple[Tensor,...], # Batch item(s)
        split_idx:int, # Train (0) or valid (1) index
        **kwargs
    ) -> Tensor|tuple[Tensor,...]:
        "Call `super().__call__` if `self.do`"
        self.before_call(b, split_idx=split_idx)
        return super().__call__(b, split_idx=split_idx, **kwargs) if self.do else b

    def _do_call(self,
        f, # Transform
        x:Tensor|tuple[Tensor,...], # Batch item(s)
        **kwargs
    ) -> Tensor|tuple[Tensor,...]:
        "Override `Transform._do_call` to apply transform `f` to `x[self.idxs]`"
        if not _is_tuple(x):
            if f is None: return x
            ret = f.returns(x) if hasattr(f,'returns') else None
            return retain_type(self._do_f(f, x, **kwargs), x, ret)
        res = tuple(self._do_call(f, x_, **kwargs) for x_ in x)
        return retain_type(res, x)

    def _do_f(self,
        f, # Transform
        x:Tensor, # Batch item
        **kwargs
    ) -> Tensor:
        "Apply transform `f` to `x[self.idxs]`"
        x[self.idxs] = f(x[self.idxs], **kwargs)
        return x



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/multiloss.ipynb.

# %% ../nbs/multiloss.ipynb 2
from __future__ import annotations
from types import FunctionType

from torch.distributions.beta import Beta

from fastai.callback.core import Callback
from fastai.callback.mixup import reduce_loss
from fastai.learner import Recorder
from fastai.layers import NoneReduce

from .basics import is_listish
from .metrics import AvgLossX, AvgSmoothLossX
from .imports import *

# %% auto 0
__all__ = ['MultiLoss', 'MultiTargetLoss', 'MixHandlerX', 'MultiLossCallback']

# %% ../nbs/multiloss.ipynb 5
def init_loss(l, **kwargs):
    "Initiatize loss class or partial loss function"
    return partialler(l, reduction='none') if isinstance(l, FunctionType) else l(reduction='none', **kwargs)

# %% ../nbs/multiloss.ipynb 6
class MultiLoss(Module):
    """
    Combine multiple `loss_funcs` on one prediction & target via `reduction`, with optional weighting.

    Log `loss_funcs` as metrics via `MultiLossCallback`, optionally using `loss_names`.
    """
    def __init__(self,
        loss_funcs:Listy[Callable[...,nn.Module]|FunctionType], # Uninitialized loss functions or classes. Must support PyTorch `reduction` string.
        weights:Listy[Numeric|Tensor]|None=None, # Weight per loss. Defaults to uniform weighting.
        loss_kwargs:Listy[dict[str,Any]]|None=None, # kwargs to pass to each loss function. Defaults to None.
        loss_names:Listy[str]|None=None, # Loss names to log using `MultiLossCallback`. Defaults to loss `__name__`.
        reduction:str|None='mean' # PyTorch loss reduction
    ):
        store_attr(but='loss_names')
        assert is_listy(loss_funcs), f"{loss_funcs=} must be list-like"

        if weights is None or len(weights)==0:
            self.weights = [1]*len(loss_funcs)
        else:
            assert len(loss_funcs) == len(weights), f"Must provide same number of {weights=} as {loss_funcs=}"
            self.weights = weights

        if loss_kwargs is None or len(loss_kwargs)==0:
            loss_kwargs = [{}]*len(loss_funcs)
        else:
            assert len(loss_funcs) == len(loss_kwargs), f"Must provide same number of {loss_kwargs=} as {loss_funcs=}"

        if loss_names is None or len(loss_names)==0:
            loss_names = [l.__name__ for l in loss_funcs]
        else:
            assert len(loss_funcs) == len(loss_names), f"Must provide same number of {loss_names=} as {loss_funcs=}"

        self.loss_funcs = [init_loss(l, **k) for l, k in zip(loss_funcs, loss_kwargs)]
        self.loss_names = L(loss_names)
        self._reduction,self._loss = reduction,{}

        for loss in self.loss_funcs:
            if getattr(loss, 'y_int', False):
                self.y_int = True

    def forward(self, pred, targ):
        for i, loss_func in enumerate(self.loss_funcs):
            l = self.weights[i]*loss_func(pred, targ)
            if i == 0: loss = torch.zeros_like(targ).float()
            loss += l
            self._loss[i] = l

        return loss.mean() if self._reduction=='mean' else loss.sum() if self._reduction=='sum' else loss

    def forward_mixup(self, pred, targ1, targ2, lam):
        for i, loss_func in enumerate(self.loss_funcs):
            l = self.weights[i]*torch.lerp(loss_func(pred, targ1), loss_func(pred, targ2), lam)
            if i == 0:
                loss = torch.zeros_like(targ1).float()
            loss += l
            self._loss[i] = l

        return loss.mean() if self._reduction=='mean' else loss.sum() if self._reduction=='sum' else loss

    @property
    def losses(self): return self._loss

    @property
    def reduction(self): return self._reduction

    @reduction.setter
    def reduction(self, r): self._reduction = r

    @delegates(Module.to)
    def to(self, *args, **kwargs):
        device, dtype, non_blocking, convert_to_format = torch._C._nn._parse_to(*args, **kwargs)
        if is_listish(self.weights) or not isinstance(self.weights, torch.Tensor):
            self.weights = torch.Tensor(self.weights)
        if self.weights.device != device:
            self.weights = self.weights.to(device=device)
        super().to(*args, **kwargs)

    def activation(self, pred):
        "Returns first `loss_funcs` `activation`"
        return getattr(self.loss_funcs[0], 'activation', noop)(pred)

    def decodes(self, pred):
        "Returns first `loss_funcs` `decodes`"
        return getattr(self.loss_funcs[0], 'decodes', noop)(pred)

# %% ../nbs/multiloss.ipynb 11
class MultiTargetLoss(MultiLoss):
    """
    Combine `loss_funcs` from multiple predictions & targets via `reduction`, with optional weighting.

    Log `loss_funcs` as metrics via `MultiLossCallback`, optionally using `loss_names`.
    """
    def __init__(self,
        loss_funcs:Listy[Callable[...,nn.Module]|FunctionType], # Uninitialized loss functions or classes. One per prediction and target. Must support PyTorch `reduction` string.
        weights:Listy[Numeric|Tensor]|None=None, # Weight per loss. Defaults to uniform weighting.
        loss_kwargs:Listy[dict[str,Any]]|None=None, # kwargs to pass to each loss function. Defaults to None.
        loss_names:Listy[str]|None=None, # Loss names to log using `MultiLossCallback`. Defaults to loss `__name__`.
        reduction:str|None='mean' # PyTorch loss reduction
    ):
        super().__init__(loss_funcs, weights, loss_kwargs, loss_names, reduction)

    def forward(self, preds, targs):
        for i, (loss_func, pred, targ) in enumerate(zip(self.loss_funcs, preds, targs)):
            l = TensorBase(self.weights[i]*loss_func(pred, targ))
            if i == 0: loss = TensorBase(torch.zeros_like(targ)).float()
            loss += l
            self._loss[i] = l

        return loss.mean() if self._reduction=='mean' else loss.sum() if self._reduction=='sum' else loss

    def forward_mixup(self, **kwargs):
        raise NotImplementedError("Mixup doesn't support Multi-Target training")

    def activation(self, preds):
        "Returns list of `activation`"
        return [getattr(self.loss_funcs[i], 'activation', noop)(pred) for i, pred in enumerate(preds)]

    def decodes(self, preds):
        "Returns list of `decodes`"
        return [getattr(self.loss_funcs[i], 'decodes', noop)(pred) for i, pred in enumerate(preds)]

# %% ../nbs/multiloss.ipynb 16
class MultiAvgLoss(AvgLossX):
    "Average the MultiLoss losses taking into account potential different batch sizes"
    def __init__(self,
        i:int, # `Multiloss` loss function location
        name:str, # Loss function name
        reduction:str|None='mean' # Override loss reduction for logging
    ):
        store_attr(but='name')
        self._name = name

    def accumulate(self, learn):
        bs = find_bs(learn.yb)
        loss = learn.loss_func.losses[self.i]
        loss = loss.mean() if self.reduction=='mean' else loss.sum() if self.reduction=='sum' else loss
        self.total += learn.to_detach(loss)*bs
        self.count += bs

# %% ../nbs/multiloss.ipynb 17
class MultiAvgSmoothLoss(AvgSmoothLossX):
    "Smooth average of the MultiLoss losses (exponentially weighted with `beta`)"
    def __init__(self,
        i:int, # `Multiloss` loss function location
        name:str, # Loss function name
        beta:float=0.98, # Smoothing beta
        reduction:str|None='mean' # Override loss reduction for logging
    ):
        super().__init__()
        store_attr(but='name')
        self._name = name

    def accumulate(self, learn):
        self.count += 1
        loss = learn.loss_func.losses[self.i]
        loss = loss.mean() if self.reduction=='mean' else loss.sum() if self.reduction=='sum' else loss
        self.val = torch.lerp(to_detach(loss, gather=False), self.val, self.beta)

# %% ../nbs/multiloss.ipynb 18
class MultiAvgSmoothLossMixup(AvgSmoothLossX):
    "Smooth average of the MultiLoss losses (exponentially weighted with `beta`)"
    def __init__(self,
        i:int, # `Multiloss` loss function location
        name:str, # Loss function name
        beta:float=0.98, # Smoothing beta
        reduction:str|None='mean' # Override loss reduction for logging
    ):
        super().__init__()
        store_attr(but='name')
        self._name = name

    def accumulate(self, learn):
        self.count += 1
        loss = learn.loss_func_mixup.losses[self.i]
        loss = loss.mean() if self.reduction=='mean' else loss.sum() if self.reduction=='sum' else loss
        self.val = torch.lerp(to_detach(loss, gather=False), self.val, self.beta)

# %% ../nbs/multiloss.ipynb 20
# MixHandlerX is based on fastai's MixHandler:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

class MixHandlerX(Callback):
    "A handler class for implementing `MixUp` style scheduling. Like fastai's `MixHandler` but supports `MultiLoss`."
    run_valid = False
    def __init__(self,
        alpha:float=0.5, # Alpha & beta parametrization for `Beta` distribution
        interp_label:bool|None=None # Blend or stack labels. Defaults to `loss_func.y_int` if None
    ):
        store_attr()
        self.distrib = Beta(tensor(alpha), tensor(alpha))

    def before_fit(self):
        "Determine whether to stack or interpolate labels"
        self.multiloss = isinstance(self.learn.loss_func, MultiLoss)
        if self.interp_label is None:
            self.stack_y = getattr(self.learn.loss_func, 'y_int', False)
        else:
            self.stack_y = not self.interp_label

    def before_train(self):
        "Determine whether to stack y"
        if self.stack_y:
            if self.multiloss:
                self.learn.loss_func_mixup = self.learn.loss_func
                self.learn.loss_func = self.multi_lf
            else:
                self.old_lf = self.learn.loss_func
                self.learn.loss_func = self.solo_lf

    def after_train(self):
        "Set the loss function back to the original loss"
        if self.stack_y:
            if self.multiloss:
                self.learn.loss_func = self.learn.loss_func_mixup
            else:
                self.learn.loss_func = self.old_lf

    def after_cancel_train(self):
        "If training is canceled, still set the loss function back"
        self.after_train()

    def after_cancel_fit(self):
        "If fit is canceled, still set the loss function back"
        self.after_train()

    def solo_lf(self, pred, *yb):
        "Interpolates losses on stacked labels by `self.lam` during training"
        if not self.training:
            return self.old_lf(pred, *yb)

        with NoneReduce(self.old_lf) as lf:
            loss = torch.lerp(lf(pred,*self.yb1), lf(pred,*yb), self.lam)

        return reduce_loss(loss, getattr(self.old_lf, 'reduction', 'mean'))

    def multi_lf(self, pred, *yb):
        if not self.training:
            return self.learn.loss_func_mixup(pred, *yb)
        else:
            return self.learn.loss_func_mixup.forward_mixup(pred, *self.yb1, *yb, self.lam)

# %% ../nbs/multiloss.ipynb 23
class MultiLossCallback(Callback):
    "Callback to automatically log and name `MultiLoss` losses as fastxtend metrics"
    run_valid,order = False,Recorder.order-1

    def __init__(self,
        beta:float=0.98, # Smoothing beta
        reduction:str|None='mean' # Override loss reduction for logging
    ):
        store_attr()

    def before_fit(self):
        if not isinstance(self.loss_func, MultiLoss):
            raise ValueError("`MultiLossCallback` requires loss to be `MultiLoss` class")

        mixup = len(self.learn._grab_cbs(MixHandlerX)) > 0 and getattr(self.learn.loss_func, 'y_int', False)

        mets= L()
        reduction = self.loss_func.reduction if self.reduction is None else self.reduction
        for i in range(len(self.loss_func.loss_funcs)):
            if mixup:
                mets += MultiAvgSmoothLossMixup(i, self.loss_func.loss_names[i], self.beta, reduction)
            else:
                mets += MultiAvgSmoothLoss(i, self.loss_func.loss_names[i], self.beta, reduction)
            mets += MultiAvgLoss(i, self.loss_func.loss_names[i], reduction)

        self.learn.metrics = mets + self.learn.metrics



from .augment import *
from .augment.all import *
from ..basics import *
from ..callback.all import *
from .data import *
from ..data.all import *
from ..losses import *
from ..metrics import *
from .models.all import *
from ..multiloss import *
from ..optimizer.all import *
from ..schedulers import *
from ..utils import *


# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/audio.01_core.ipynb.

# %% ../../nbs/audio.01_core.ipynb 1
# Contains code from https://github.com/fastaudio/fastaudio:
# fastaudio - MIT License - Copyright (c) 2020 scart97

# %% ../../nbs/audio.01_core.ipynb 4
from __future__ import annotations

import torchaudio
import inspect

from IPython.display import Audio, display
from librosa.display import waveshow, specshow
from librosa import amplitude_to_db

from fastai.imports import Path, mimetypes, plt
from fastai.torch_core import TensorBase, TensorImageBase

from ..imports import *

# %% auto 0
__all__ = ['audio_extensions', 'show_audio_signal', 'TensorAudio', 'show_spectrogram', 'TensorSpec', 'TensorMelSpec']

# %% ../../nbs/audio.01_core.ipynb 6
audio_extensions = tuple(
    str.lower(k) for k, v in mimetypes.types_map.items() if v.startswith("audio/")
)

# %% ../../nbs/audio.01_core.ipynb 8
def show_audio_signal(at, ctx, ax=None, title="", sep=0.03, **kwargs):
    ax = ifnone(ax, ctx)
    if ax is None:
        _, ax = plt.subplots()
    ax.axis(False)
    step = 0
    if at.channels==1: sep=0
    for i, channel in enumerate(at):
        step += 1 / (at.channels)
        top = sep if i==0 else 0
        ia = ax.inset_axes((0, 1-(step-top), 1, (1 / at.channels)-sep))
        waveshow(channel.cpu().numpy(), sr=at.sr, ax=ia, x_axis='time' if i==at.channels-1 else 'none', **kwargs)
    ax.set_title(title)

    return ax

# %% ../../nbs/audio.01_core.ipynb 9
class TensorAudio(TensorBase):
    "Tensor for audio. Can be created from files and has extra properties. Also knows how to show itself."

    @classmethod
    @delegates(torchaudio.load, keep=True)
    def create(cls, fn, **kwargs):
        "Creates `TensorAudio` from file `fn`"
        sig, sr = torchaudio.load(fn, **kwargs)
        return cls(sig, sr=sr)

    def __new__(cls, x, sr=None, **kwargs):
        return super().__new__(cls, x, sr=sr, **kwargs)

    @property
    def samples(self):
        return self.shape[-1]

    @property
    def channels(self):
        return self.shape[-2]

    @property
    def duration(self):
        return self.samples / float(self.sr)

    def hear(self):
        "Listen to audio clip. Creates a html player."
        display(Audio(self.cpu(), rate=self.sr))

    @delegates(show_audio_signal)
    def show(self, ctx=None, hear=True, **kwargs):
        "Show audio clip using librosa. Pass `hear=True` to also display a html player to listen."
        if hear: self.hear()
        return show_audio_signal(self, ctx=ctx, **kwargs)

    def save(self, fn: Path, overwrite=True):
        "Save the audio into the specfied path"
        fn = Path.expanduser(fn)
        if not overwrite and Path.exists(fn):
            raise Exception("File already exists")
        torchaudio.save(fn, self.data, self.sr)

# %% ../../nbs/audio.01_core.ipynb 17
def show_spectrogram(aspec, title="", ax=None, ctx=None, sep=0.025, to_db=False, **kwargs):
    ax = ifnone(ax, ctx)
    if ax is None:
        _, ax = plt.subplots()
    ax.axis(False)
    step = 0
    if aspec.channels==1: sep=0
    for i, channel in enumerate(aspec):
        step += 1 / (aspec.channels)
        top = sep if i==0 else 0
        ia = ax.inset_axes((0, 1-(step-top), 1, (1 / aspec.channels)-sep))
        if to_db: channel = amplitude_to_db(channel.cpu().numpy())
        else:     channel = channel.cpu().numpy()
        z = specshow(channel, ax=ia, **aspec._all_show_args(), x_axis='time' if i==aspec.channels-1 else None, **kwargs)
        if i == 0:  # Only colorbar the first one
            plt.colorbar(z, format=aspec._colorbar_fmt, ax=ax)
    ax.set_title(title)

    return ax

# %% ../../nbs/audio.01_core.ipynb 18
def get_usable_kwargs(func, kwargs, exclude=None):
    exclude = ifnone(exclude, [])
    defaults = {
        k: v.default
        for k, v in inspect.signature(func).parameters.items()
        if k not in exclude
    }
    usable = {k: v for k, v in kwargs.items() if k in defaults}
    return {**defaults, **usable}

# %% ../../nbs/audio.01_core.ipynb 19
class TensorSpec(TensorImageBase):
    "Tensor for Audio Spectrograms. Has extra properties and knows how to show itself."

    @classmethod
    def create(cls,
        ta:Tensor,
        settings:dict|None=None
    ):
        "Create an TensorSpec from a torch tensor"
        spec = cls(ta)
        spec._settings = settings
        return spec

    @property
    def duration(self):
        # spectrograms round up length to fill incomplete columns, so subtract 0.5 to compensate
        return (self.hop_length * (self.shape[-1] - 0.5)) / self.sr

    @property
    def width(self):
        return self.shape[-1]

    @property
    def height(self):
        return self.shape[-2]

    @property
    def channels(self):
        return self.shape[-3]

    def _all_show_args(self):
        proper_kwargs = get_usable_kwargs(specshow, self._settings,
                                          exclude=["ax", "kwargs", "x_axis", "data"])
        proper_kwargs.update({"y_axis": "log"})
        proper_kwargs.update(self._show_args)
        return proper_kwargs

    @property
    def _colorbar_fmt(self):
        return "%+2.0f dB" if "to_db" in self._settings and self.to_db else "%+2.0f"

    def __getattr__(self, name):
        if name == "settings":
            return self._settings
        if not name.startswith("_"):
            return self._settings[name]
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {name}"
        )

    @delegates(show_spectrogram)
    def show(self, ctx=None, ax=None, title="", **kwargs):
        "Show spectrogram using librosa"
        return show_spectrogram(self, ctx=ctx, ax=ax, title=title, **kwargs)

# %% ../../nbs/audio.01_core.ipynb 27
class TensorMelSpec(TensorSpec):
    "Tensor for Audio MelSpectrograms. Has extra properties and knows how to show itself."

    @classmethod
    def create(cls,
        ta:Tensor,
        settings:dict|None=None
    ):
        "Create an TensorMelSpec from a torch tensor"
        mel = cls(ta)
        mel._settings = settings
        return mel

    def _all_show_args(self):
        proper_kwargs = get_usable_kwargs(specshow, self._settings,
                                          exclude=["ax", "kwargs", "x_axis", "data"])
        proper_kwargs.update({"y_axis": "mel"})
        proper_kwargs.update(self._show_args)
        return proper_kwargs

    @property
    def duration(self):
        # spectrograms round up length to fill incomplete columns, so subtract 0.5 to compensate
        return (self.hop_length * (self.shape[-1] - 0.5)) / self.sr

    @property
    def width(self):
        return self.shape[-1]

    @property
    def height(self):
        return self.shape[-2]

    @property
    def channels(self):
        return self.shape[-3]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/callback.cutmixup.ipynb.

# %% ../../nbs/callback.cutmixup.ipynb 1
# Contains code from:
# fastai - Apache License 2.0 - Copyright (c) 2023 fast.ai

# %% ../../nbs/callback.cutmixup.ipynb 4
from __future__ import annotations

from torch.distributions import Categorical
from torch.distributions.beta import Beta

from fastcore.transform import Pipeline, Transform

from fastai.data.transforms import IntToFloatTensor, Normalize
from fastai.callback.mixup import reduce_loss
from fastai.layers import NoneReduce
from fastai.vision.augment import AffineCoordTfm, RandomResizedCropGPU

from ..multiloss import MixHandlerX
from ..imports import *

# %% auto 0
__all__ = ['MixUp', 'CutMix', 'CutMixUp', 'CutMixUpAugment']

# %% ../../nbs/callback.cutmixup.ipynb 6
class MixUp(MixHandlerX):
    "Implementation of https://arxiv.org/abs/1710.09412. Supports `MultiLoss`"
    def __init__(self,
        alpha:float=.4, # Alpha & beta parametrization for `Beta` distribution
        interp_label:bool|None=None # Blend or stack labels. Defaults to loss' `y_int` if None
    ):
        super().__init__(alpha, interp_label)

    def before_batch(self):
        "Blend inputs and labels with another random item in the batch"
        shuffle = torch.randperm(self.y.size(0)).to(self.x.device)
        xb1,self.yb1 = tuple(L(self.xb).itemgot(shuffle)),tuple(L(self.yb).itemgot(shuffle))
        self.lam = self._mixup(self.y.size(0))
        nx_dims = len(self.x.size())
        self.learn.xb = tuple(L(xb1,self.xb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=nx_dims-1)))

        if not self.stack_y:
            ny_dims = len(self.y.size())
            self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))

    def _mixup(self, bs):
        lam = self.distrib.sample((bs,)).to(self.x.device)
        if len(lam.shape) > 1:
            lam = lam.squeeze()
        lam = torch.stack([lam, 1-lam], 1)
        return lam.max(1)[0]

# %% ../../nbs/callback.cutmixup.ipynb 8
class CutMix(MixHandlerX):
    "Implementation of https://arxiv.org/abs/1905.04899. Supports `MultiLoss`"
    def __init__(self, 
        alpha:float=1., # Alpha & beta parametrization for `Beta` distribution
        uniform:bool=True, # Uniform patches across batch. True matches fastai CutMix
        interp_label:bool|None=None # Blend or stack labels. Defaults to loss' `y_int` if None
    ): 
        super().__init__(alpha, interp_label)
        store_attr(but='alpha,interp_label')

    def before_batch(self):
        "Add patches and blend labels from another random item in batch"
        bs, _, H, W = self.x.size()
        shuffle = torch.randperm(bs).to(self.x.device)
        xb1,self.yb1 = self.x[shuffle], (self.y[shuffle],)

        if self.uniform:
            xb, self.lam = self._uniform_cutmix(self.x, xb1, H, W)
        else:
            xb, self.lam = self._multi_cutmix(self.x, xb1, H, W, bs)
        self.learn.xb = (xb,)

        if not self.stack_y:
            ny_dims = len(self.y.size())
            self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))

    def _uniform_cutmix(self, xb, xb1, H, W):
        "Add uniform patches and blend labels from another random item in batch"
        lam = self.distrib.sample((1,)).to(self.x.device)
        x1, y1, x2, y2 = self.rand_bbox(W, H, lam)
        xb[..., y1:y2, x1:x2] = xb1[..., y1:y2, x1:x2]
        lam = (1 - ((x2-x1)*(y2-y1))/float(W*H))
        return xb, lam

    def _multi_cutmix(self, xb, xb1, H, W, bs):
        "Add random patches and blend labels from another random item in batch"
        lam = self.distrib.sample((bs,)).to(self.x.device)
        for i in range(bs):
            if 1 > lam[i] > 0:
                x1, y1, x2, y2 = self.rand_bbox(W, H, lam[i])
                xb[i, ..., y1:y2, x1:x2] = xb1[i, ..., y1:y2, x1:x2]
                lam[i] = (1 - ((x2-x1)*(y2-y1))/float(W*H))
        return xb, lam

    def rand_bbox(self,
        W:int, # Input image width
        H:int, # Input image height
        lam:Tensor # Lambda sample from Beta distribution
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]: # Top-left and bottom-right coordinates
        "Return random sub coordinates"
        cut_rat = torch.sqrt(1. - lam).to(self.x.device)
        cut_w = torch.round(W * cut_rat).type(torch.long).to(self.x.device)
        cut_h = torch.round(H * cut_rat).type(torch.long).to(self.x.device)
        # uniform
        cx = torch.randint(0, W, (1,)).to(self.x.device)
        cy = torch.randint(0, H, (1,)).to(self.x.device)
        x1 = torch.clamp(cx - torch.div(cut_w, 2, rounding_mode='floor'), 0, W)
        y1 = torch.clamp(cy - torch.div(cut_h, 2, rounding_mode='floor'), 0, H)
        x2 = torch.clamp(cx + torch.div(cut_w, 2, rounding_mode='floor'), 0, W)
        y2 = torch.clamp(cy + torch.div(cut_h, 2, rounding_mode='floor'), 0, H)
        return x1, y1, x2, y2

# %% ../../nbs/callback.cutmixup.ipynb 10
class CutMixUp(MixUp, CutMix):
    """
    Combo implementation of https://arxiv.org/abs/1710.09412 and https://arxiv.org/abs/1905.04899

    Supports element-wise application of MixUp and CutMix on a batch.
    """
    run_valid = False
    def __init__(self,
        mix_alpha:float=.4, # MixUp alpha & beta parametrization for `Beta` distribution
        cut_alpha:float=1., # CutMix alpha & beta parametrization for `Beta` distribution
        mixup_ratio:Numeric=1, # Ratio to apply `MixUp` relative to `CutMix`
        cutmix_ratio:Numeric=1, # Ratio to apply `CutMix` relative to `MixUp`
        cutmix_uniform:bool=True, # Uniform patches across batch. True matches fastai CutMix
        element:bool=True, # Apply element-wise MixUp and CutMix on a batch
        interp_label:bool|None=None # Blend or stack labels. Defaults to loss' `y_int` if None
    ):
        store_attr()
        if element:
            total = mixup_ratio + cutmix_ratio
            self.categorical = Categorical(tensor([mixup_ratio/total, cutmix_ratio/total]))
        MixUp.__init__(self, mix_alpha, interp_label)
        CutMix.__init__(self, cut_alpha, cutmix_uniform, interp_label)
        self.mix_distrib = Beta(tensor(mix_alpha), tensor(mix_alpha))
        self.cut_distrib = Beta(tensor(cut_alpha), tensor(cut_alpha))
        self.ratio = mixup_ratio / (cutmix_ratio + mixup_ratio)

    def before_batch(self):
        "Apply MixUp or CutMix"
        if self.element:
            xb, self.yb1 = self.x, self.y
            bs, _, _, _ = xb.size()
            self.lam = torch.zeros(bs, device=xb.device)
            aug_type = self.categorical.sample((bs,))
            shuffle = torch.randperm(xb.shape[0]).to(xb.device)
            xb1, self.yb1 = xb[shuffle], self.yb1[shuffle]

            # Apply MixUp
            if (aug_type==0).sum() > 0:
                self.distrib = self.mix_distrib
                self.lam[aug_type==0] = MixUp._mixup(self, xb[aug_type==0].shape[0])
                xb[aug_type==0] = torch.lerp(xb1[aug_type==0], xb[aug_type==0], weight=unsqueeze(self.lam[aug_type==0], n=3))

            # Apply CutMix
            bs, _, H, W = xb[aug_type==1].size()
            if bs > 0:
                self.distrib = self.cut_distrib
                if self.cutmix_uniform:
                    xb[aug_type==1], self.lam[aug_type==1] = CutMix._uniform_cutmix(self, xb[aug_type==1], xb1[aug_type==1], H, W)
                else:
                    xb[aug_type==1], self.lam[aug_type==1] = CutMix._multi_cutmix(self, xb[aug_type==1], xb1[aug_type==1], H, W, bs)

            self.learn.xb = (xb,)
            if not self.stack_y:
                ny_dims = len(self.yb1.size())
                self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))
            else:
                self.yb1 = (self.yb1,)

        elif torch.rand(1) <= self.ratio: #mixup
            self.distrib = self.mix_distrib
            MixUp.before_batch(self)
        else:
            self.distrib = self.cut_distrib
            CutMix.before_batch(self)

# %% ../../nbs/callback.cutmixup.ipynb 12
def _do_cutmixaug(t:Tensor):
    return t.sum().item() > 0

# %% ../../nbs/callback.cutmixup.ipynb 13
class CutMixUpAugment(MixUp, CutMix):
    """
    Combo implementation of https://arxiv.org/abs/1710.09412 and https://arxiv.org/abs/1905.04899 plus Augmentation.

    Supports element-wise application of MixUp, CutMix, and Augmentation on a batch.

    Pulls augmentations from `Dataloaders.train.after_batch`. These augmentations are not applied when performing `MixUp` & `CutMix`, the frequency is controlled by `augment_ratio`.

    Use `augment_finetune` to only apply dataloader augmentations at the end of training.

    `cutmixup_augs` are an optional separate set of augmentations to apply with `MixUp` and `CutMix`. Usually less intensive then the dataloader augmentations. 
    """
    run_valid = False
    def __init__(self,
        mix_alpha:float=.4, # MixUp alpha & beta parametrization for `Beta` distribution
        cut_alpha:float=1., # CutMix alpha & beta parametrization for `Beta` distribution
        mixup_ratio:Numeric=1, # Ratio to apply `MixUp` relative to `CutMix` & augmentations
        cutmix_ratio:Numeric=1, # Ratio to apply `CutMix` relative to `MixUp` & augmentations
        augment_ratio:Numeric=1, # Ratio to apply augmentations relative to `MixUp` & `CutMix`
        augment_finetune:Numeric|None=None, # Number of epochs or pct of training to only apply augmentations
        cutmix_uniform:bool=True, # Uniform patches across batch. True matches fastai CutMix
        cutmixup_augs:Listified[Transform|Callable[...,Transform]]|None=None, # Augmentations to apply before `MixUp` & `CutMix`. Should not have `Normalize`
        element:bool=True, # Apply element-wise MixUp, CutMix, and Augment on a batch
        interp_label:bool|None=None, # Blend or stack labels. Defaults to loss' `y_int` if None
    ):
        store_attr()
        if element:
            total = mixup_ratio + cutmix_ratio + augment_ratio
            self.categorical = Categorical(tensor([mixup_ratio/total, cutmix_ratio/total, augment_ratio/total]))
        MixUp.__init__(self, mix_alpha, interp_label)
        CutMix.__init__(self, cut_alpha, cutmix_uniform, interp_label)
        self.mix_distrib = Beta(tensor(mix_alpha), tensor(mix_alpha))
        self.cut_distrib = Beta(tensor(cut_alpha), tensor(cut_alpha))
        self.aug_cutmix_ratio = augment_ratio / (augment_ratio + cutmix_ratio + mixup_ratio)
        if self.aug_cutmix_ratio == 1: self.cut_mix_ratio = 0
        else: self.cut_mix_ratio = mixup_ratio / (cutmix_ratio + mixup_ratio)
        self._docutmixaug = cutmixup_augs is not None

    def before_fit(self):
        "Remove training augmentations from dataloader & setup augmentation pipelines"
        super().before_fit()
        if self.augment_finetune is None:
            self.augment_finetune = (self.learn.n_epoch + 1)/self.learn.n_epoch
        elif self.augment_finetune >= 1:
            self.augment_finetune = self.augment_finetune/self.learn.n_epoch
        else:
            self.augment_finetune = self.augment_finetune

        self._inttofloat_pipe = Pipeline([])
        self._norm_pipe = Pipeline([])
        if self._docutmixaug:
            self._cutmixaugs_pipe = Pipeline(self.cutmixup_augs)

        # first copy transforms
        self._orig_pipe = self.dls.train.after_batch
        self._orig_pipe.split_idx = 0 # need to manually set split_idx for training augmentations to run

        # Loop through existing transforms looking for IntToFloatTensor, Normalize
        self._size, mode, augs = None, None, []
        for aug in self.dls.train.after_batch.fs:
            if isinstance(aug, IntToFloatTensor):
                self._inttofloat_pipe = Pipeline([aug])
            else:
                if isinstance(aug, Normalize):
                    self._norm_pipe = Pipeline([aug])
                elif isinstance(aug, (AffineCoordTfm, RandomResizedCropGPU)) and aug.size is not None:
                    self._size = aug.size
                    mode = aug.mode
                if self.element:
                    augs.append(aug)

        # One Batch requires IntToFloatTensor before self._aug_pipe is called
        if self.element: self._aug_pipe = Pipeline(augs)

        # If there is a resize in Augmentations and no `cutmixup_augs`, need to replicate it for MixUp/CutMix
        if not self._docutmixaug and self._size is not None:
            self._docutmixaug = True
            self._cutmixaugs_pipe = Pipeline([AffineCoordTfm(size=self._size, mode=mode)])

        # set existing transforms to an empty Pipeline
        self.dls.train.after_batch = Pipeline([])

    def before_batch(self):
        "Apply MixUp, CutMix, optional MixUp/CutMix augmentations, and/or augmentations"
        if self.element and self.augment_finetune >= self.learn.pct_train:
            self._doaugs = False
            xb, self.yb1 = self.x, self.y
            bs, C, H, W = xb.size()
            self.lam = torch.zeros(bs, device=xb.device)
            aug_type = self.categorical.sample((bs,))
            do_mix, do_cut, do_aug = _do_cutmixaug(aug_type==0), _do_cutmixaug(aug_type==1), _do_cutmixaug(aug_type==2)
            if do_mix or do_cut:
                shuffle = torch.randperm(xb[aug_type<2].shape[0]).to(xb.device)
                self.yb1[aug_type<2] = self.yb1[aug_type<2][shuffle]

            # Apply IntToFloat to all samples
            xb = self._inttofloat_pipe(xb)

            # New Tensor for possibly resized
            xb2 = torch.zeros([bs, C, self._size[0], self._size[1]], dtype=xb.dtype, device=xb.device) if self._size is not None else torch.zeros_like(xb)

            # Apply MixUp/CutMix Augmentations to MixUp and CutMix samples
            if do_mix or do_cut:
                if self._docutmixaug:
                    xb2[aug_type<2] = self._cutmixaugs_pipe(xb[aug_type<2])
                else:
                    xb2[aug_type<2] = xb[aug_type<2]

            # Original Augmentations
            if do_aug:
                xb2[aug_type==2] = self._aug_pipe(xb[aug_type==2])

            # Possibly Resized xb and shuffled xb1
            xb = xb2
            if do_mix or do_cut:
                xb1 = xb[aug_type<2][shuffle]

            # Apply MixUp
            if do_mix:
                self.distrib = self.mix_distrib
                self.lam[aug_type==0] = MixUp._mixup(self, xb[aug_type==0].shape[0])
                xb[aug_type==0] = torch.lerp(xb1[aug_type[aug_type<2]==0], xb[aug_type==0], weight=unsqueeze(self.lam[aug_type==0], n=3))

            # Apply CutMix
            if do_cut:
                bs, _, H, W = xb[aug_type==1].size()
                self.distrib = self.cut_distrib
                if self.cutmix_uniform:
                    xb[aug_type==1], lam = CutMix._uniform_cutmix(self, xb[aug_type==1], xb1[aug_type[aug_type<2]==1], H, W)
                    self.lam[aug_type==1] = lam.expand(bs)
                else:
                    xb[aug_type==1], self.lam[aug_type==1] = CutMix._multi_cutmix(self, xb[aug_type==1], xb1[aug_type[aug_type<2]==1], H, W, bs)

            # Normalize MixUp and CutMix
            if do_mix or do_cut:
                xb[aug_type<2] = self._norm_pipe(xb[aug_type<2])

            self.learn.xb = (xb,)
            if not self.stack_y:
                ny_dims = len(self.yb1.size())
                self.learn.yb = tuple(L(self.yb1,self.yb).map_zip(torch.lerp,weight=unsqueeze(self.lam, n=ny_dims-1)))
            else:
                self.yb1 = (self.yb1,)

        elif self.augment_finetune >= self.learn.pct_train and torch.rand(1) >= self.aug_cutmix_ratio: # augs or mixup/cutmix
            self._doaugs = False

            # Apply IntToFloat to MixUp/CutMix and MixUp/CutMix Augmentations
            self.learn.xb = self._inttofloat_pipe(self.xb)
            if self._docutmixaug:
                self.learn.xb = self._cutmixaugs_pipe(self.xb)

            # Perform MixUp or CutMix
            if self.cut_mix_ratio > 0 and torch.rand(1) <= self.cut_mix_ratio:
                self.distrib = self.mix_distrib
                MixUp.before_batch(self)
            else:
                self.distrib = self.cut_distrib
                CutMix.before_batch(self)

            # Normalize MixUp/CutMix
            self.learn.xb = self._norm_pipe(self.xb) # now normalize
        else:
            # Original Augmentations
            self._doaugs = True
            self.learn.xb = self._orig_pipe(self.xb)

    def after_fit(self):
        "Reset the train dataloader augmentations"
        self.dls.train.after_batch = self._orig_pipe

    def after_cancel_fit(self):
        "Reset the train dataloader augmentations and loss function"
        self.after_fit()
        MixUp.after_cancel_fit(self)

    def solo_lf(self, pred, *yb):
        "`norm_lf` applies the original loss function on both outputs based on `self.lam` if applicable"
        if not self.training or self._doaugs:
            return self.old_lf(pred, *yb)
        with NoneReduce(self.old_lf) as lf:
            loss = torch.lerp(lf(pred,*self.yb1), lf(pred,*yb), self.lam)
        return reduce_loss(loss, getattr(self.old_lf, 'reduction', 'mean'))

    def multi_lf(self, pred, *yb):
        "`norm_lf` applies the original loss function on both outputs based on `self.lam` if applicable"
        if not self.training or self._doaugs:
            return self.learn.loss_func_mixup(pred, *yb)
        else:
            return self.learn.loss_func_mixup.forward_mixup(pred, *self.yb1, *yb, self.lam)



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"is_listish": "0_basics.ipynb",
         "listify_store_attr": "0_basics.ipynb",
         "audio_extensions": "audio.01_core.ipynb",
         "show_audio_signal": "audio.01_core.ipynb",
         "TensorAudio": "audio.01_core.ipynb",
         "show_spectrogram": "audio.01_core.ipynb",
         "get_usable_kwargs": "audio.01_core.ipynb",
         "TensorSpec": "audio.01_core.ipynb",
         "TensorMelSpec": "audio.01_core.ipynb",
         "Spectrogram": "audio.02_data.ipynb",
         "MelSpectrogram": "audio.02_data.ipynb",
         "AudioBlock": "audio.02_data.ipynb",
         "SpecBlock": "audio.02_data.ipynb",
         "MelSpecBlock": "audio.02_data.ipynb",
         "Flip": "audio.03_augment.ipynb",
         "Roll": "audio.03_augment.ipynb",
         "AudioPadMode": "audio.03_augment.ipynb",
         "TensorAudio.crop_pad": "audio.03_augment.ipynb",
         "RandomCropPad": "audio.03_augment.ipynb",
         "VolumeMode": "audio.03_augment.ipynb",
         "Volume": "audio.03_augment.ipynb",
         "PeakNorm": "audio.03_augment.ipynb",
         "VolumeOrPeakNorm": "audio.03_augment.ipynb",
         "NoiseColor": "audio.03_augment.ipynb",
         "Noise": "audio.03_augment.ipynb",
         "VolumeBatch": "audio.03_augment.ipynb",
         "TensorAudio.pitch_shift": "audio.03_augment.ipynb",
         "PitchShift": "audio.03_augment.ipynb",
         "PitchShiftTA": "audio.03_augment.ipynb",
         "TensorAudio.time_stretch": "audio.03_augment.ipynb",
         "TimeStretch": "audio.03_augment.ipynb",
         "PitchShiftOrTimeStretch": "audio.03_augment.ipynb",
         "TimeMasking": "audio.03_augment.ipynb",
         "FrequencyMasking": "audio.03_augment.ipynb",
         "AmplitudeToDBMode": "audio.03_augment.ipynb",
         "AmplitudeToDB": "audio.03_augment.ipynb",
         "AudioNormalize": "audio.03_augment.ipynb",
         "StackSpecCallback": "audio.04_learner.ipynb",
         "audio_learner": "audio.04_learner.ipynb",
         "AudioMixHandler": "audio.05_mixup.ipynb",
         "AudioMixUp": "audio.05_mixup.ipynb",
         "AudioCutMix": "audio.05_mixup.ipynb",
         "AudioCutMixUp": "audio.05_mixup.ipynb",
         "AudioCutMixUpAugment": "audio.05_mixup.ipynb",
         "CastToTensorBackport": "callback.casttotensor.ipynb",
         "ChannelsLast": "callback.channelslast.ipynb",
         "Learner.to_channelslast": "callback.channelslast.ipynb",
         "Learner.to_contiguous": "callback.channelslast.ipynb",
         "MixUp": "callback.cutmixup.ipynb",
         "CutMix": "callback.cutmixup.ipynb",
         "CutMixUp": "callback.cutmixup.ipynb",
         "CutMixUpAugment": "callback.cutmixup.ipynb",
         "EMACallback": "callback.ema.ipynb",
         "LRFinder": "callback.lr_finder.ipynb",
         "Learner.lr_find": "callback.lr_finder.ipynb",
         "IncreaseMode": "callback.progresize.ipynb",
         "ProgressiveResize": "callback.progresize.ipynb",
         "Callback.__call__": "callback.simpleprofiler.ipynb",
         "Learner.all_batches": "callback.simpleprofiler.ipynb",
         "Learner.show_training_loop": "callback.simpleprofiler.ipynb",
         "SimpleProfilerPostCallback": "callback.simpleprofiler.ipynb",
         "SimpleProfilerCallback": "callback.simpleprofiler.ipynb",
         "Learner.profile": "callback.simpleprofiler.ipynb",
         "TerminateOnTrainNaN": "callback.tracker.ipynb",
         "SaveModelAtEnd": "callback.tracker.ipynb",
         "KFoldColSplitter": "data.transforms.ipynb",
         "ParentSplitter": "data.transforms.ipynb",
         "GreatGrandparentSplitter": "data.transforms.ipynb",
         "BCEWithLogitsLoss": "losses.ipynb",
         "ClassBalanced": "losses.ipynb",
         "ClassBalancedCrossEntropyLoss": "losses.ipynb",
         "ClassBalancedBCEWithLogitsLoss": "losses.ipynb",
         "LogMetric": "metrics.ipynb",
         "MetricType": "metrics.ipynb",
         "ActivationType": "metrics.ipynb",
         "MetricX": "metrics.ipynb",
         "AvgMetricX": "metrics.ipynb",
         "AccumMetricX": "metrics.ipynb",
         "AvgSmoothMetricX": "metrics.ipynb",
         "AvgLossX": "metrics.ipynb",
         "AvgSmoothLossX": "metrics.ipynb",
         "ValueMetricX": "metrics.ipynb",
         "Recorder.__init__": "metrics.ipynb",
         "Recorder.before_fit": "metrics.ipynb",
         "Recorder.after_batch": "metrics.ipynb",
         "Recorder.before_epoch": "metrics.ipynb",
         "Recorder.before_train": "metrics.ipynb",
         "Recorder.before_validate": "metrics.ipynb",
         "Recorder.after_train": "metrics.ipynb",
         "Recorder.after_validate": "metrics.ipynb",
         "Recorder.after_epoch": "metrics.ipynb",
         "Recorder.train_mets": "metrics.ipynb",
         "Recorder.valid_mets": "metrics.ipynb",
         "func_to_metric": "metrics.ipynb",
         "skm_to_fastxtend": "metrics.ipynb",
         "accuracy": "metrics.ipynb",
         "Accuracy": "metrics.ipynb",
         "error_rate": "metrics.ipynb",
         "ErrorRate": "metrics.ipynb",
         "top_k_accuracy": "metrics.ipynb",
         "TopKAccuracy": "metrics.ipynb",
         "APScoreBinary": "metrics.ipynb",
         "BalancedAccuracy": "metrics.ipynb",
         "BrierScore": "metrics.ipynb",
         "CohenKappa": "metrics.ipynb",
         "F1Score": "metrics.ipynb",
         "FBeta": "metrics.ipynb",
         "HammingLoss": "metrics.ipynb",
         "Jaccard": "metrics.ipynb",
         "Precision": "metrics.ipynb",
         "Recall": "metrics.ipynb",
         "RocAuc": "metrics.ipynb",
         "RocAucBinary": "metrics.ipynb",
         "MatthewsCorrCoef": "metrics.ipynb",
         "accuracy_multi": "metrics.ipynb",
         "AccuracyMulti": "metrics.ipynb",
         "APScoreMulti": "metrics.ipynb",
         "BrierScoreMulti": "metrics.ipynb",
         "F1ScoreMulti": "metrics.ipynb",
         "FBetaMulti": "metrics.ipynb",
         "HammingLossMulti": "metrics.ipynb",
         "JaccardMulti": "metrics.ipynb",
         "MatthewsCorrCoefMulti": "metrics.ipynb",
         "PrecisionMulti": "metrics.ipynb",
         "RecallMulti": "metrics.ipynb",
         "RocAucMulti": "metrics.ipynb",
         "mse": "metrics.ipynb",
         "MSE": "metrics.ipynb",
         "rmse": "metrics.ipynb",
         "RMSE": "metrics.ipynb",
         "mae": "metrics.ipynb",
         "MAE": "metrics.ipynb",
         "msle": "metrics.ipynb",
         "MSLE": "metrics.ipynb",
         "exp_rmspe": "metrics.ipynb",
         "ExpRMSE": "metrics.ipynb",
         "ExplainedVariance": "metrics.ipynb",
         "R2Score": "metrics.ipynb",
         "PearsonCorrCoef": "metrics.ipynb",
         "SpearmanCorrCoef": "metrics.ipynb",
         "foreground_acc": "metrics.ipynb",
         "ForegroundAcc": "metrics.ipynb",
         "Dice": "metrics.ipynb",
         "DiceMulti": "metrics.ipynb",
         "JaccardCoeff": "metrics.ipynb",
         "CorpusBLEUMetric": "metrics.ipynb",
         "Perplexity": "metrics.ipynb",
         "perplexity": "metrics.ipynb",
         "LossMetric": "metrics.ipynb",
         "LossMetrics": "metrics.ipynb",
         "init_loss": "multiloss.ipynb",
         "MultiLoss": "multiloss.ipynb",
         "MultiTargetLoss": "multiloss.ipynb",
         "MultiAvgLoss": "multiloss.ipynb",
         "MultiAvgSmoothLoss": "multiloss.ipynb",
         "MultiAvgSmoothLossMixup": "multiloss.ipynb",
         "MixHandlerX": "multiloss.ipynb",
         "MultiLossCallback": "multiloss.ipynb",
         "debias": "optimizers.adan.ipynb",
         "adan_setup": "optimizers.adan.ipynb",
         "adan_avgs": "optimizers.adan.ipynb",
         "adan_avgs.defaults": "optimizers.adan.ipynb",
         "adan_step": "optimizers.adan.ipynb",
         "Adan": "optimizers.adan.ipynb",
         "AdanLargeBatchLR": "optimizers.adan.ipynb",
         "FineTuneOpt": "optimizers.finetune.ipynb",
         "fine_tune_wd": "optimizers.finetune.ipynb",
         "SGDFT": "optimizers.finetune.ipynb",
         "RMSPropFT": "optimizers.finetune.ipynb",
         "AdamFT": "optimizers.finetune.ipynb",
         "RAdamFT": "optimizers.finetune.ipynb",
         "QHAdamFT": "optimizers.finetune.ipynb",
         "LarcFT": "optimizers.finetune.ipynb",
         "LambFT": "optimizers.finetune.ipynb",
         "LookaheadFT": "optimizers.finetune.ipynb",
         "rangerFT": "optimizers.finetune.ipynb",
         "Learner.fit_flat_varied": "schedulers.fit_flat_varied.ipynb",
         "TEST_AUDIO": "test_utils.ipynb",
         "TfmdDL.to": "transform.ipynb",
         "BatchRandTransform": "transform.ipynb",
         "free_gpu_memory": "utils.ipynb",
         "less_random": "utils.ipynb",
         "scale_time": "utils.ipynb",
         "GrayScaleMode": "vision.augment.batch.ipynb",
         "GrayScale": "vision.augment.batch.ipynb",
         "ChannelDrop": "vision.augment.batch.ipynb",
         "RandomNoise": "vision.augment.batch.ipynb",
         "RandomErasingBatch": "vision.augment.batch.ipynb",
         "TensorImage.resize": "vision.augment.itemtensor.ipynb",
         "TensorMask.resize": "vision.augment.itemtensor.ipynb",
         "TensorImage.crop_pad": "vision.augment.itemtensor.ipynb",
         "TensorMask.crop_pad": "vision.augment.itemtensor.ipynb",
         "encodes": "vision.augment.itemtensor.ipynb",
         "PreBatchAsItem": "vision.data.ipynb",
         "PostBatchAsItem": "vision.data.ipynb",
         "ImageCPUBlock": "vision.data.ipynb",
         "MaskCPUBlock": "vision.data.ipynb",
         "ECA": "vision.models.attention_modules.ipynb",
         "ShuffleAttention": "vision.models.attention_modules.ipynb",
         "ZPool": "vision.models.attention_modules.ipynb",
         "AttentionGate": "vision.models.attention_modules.ipynb",
         "TripletAttention": "vision.models.attention_modules.ipynb",
         "BlurPool": "vision.models.pooling.ipynb",
         "MaxBlurPool": "vision.models.pooling.ipynb",
         "ResBlock": "vision.models.xresnet.ipynb",
         "ResNeXtBlock": "vision.models.xresnet.ipynb",
         "SEBlock": "vision.models.xresnet.ipynb",
         "SEResNeXtBlock": "vision.models.xresnet.ipynb",
         "ECABlock": "vision.models.xresnet.ipynb",
         "ECAResNeXtBlock": "vision.models.xresnet.ipynb",
         "SABlock": "vision.models.xresnet.ipynb",
         "SAResNeXtBlock": "vision.models.xresnet.ipynb",
         "TABlock": "vision.models.xresnet.ipynb",
         "TAResNeXtBlock": "vision.models.xresnet.ipynb",
         "XResNet": "vision.models.xresnet.ipynb",
         "xresnet18": "vision.models.xresnet.ipynb",
         "xresnet34": "vision.models.xresnet.ipynb",
         "xresnet50": "vision.models.xresnet.ipynb",
         "xresnet101": "vision.models.xresnet.ipynb",
         "xresnext18": "vision.models.xresnet.ipynb",
         "xresnext34": "vision.models.xresnet.ipynb",
         "xresnext50": "vision.models.xresnet.ipynb",
         "xresnext101": "vision.models.xresnet.ipynb",
         "xse_resnet18": "vision.models.xresnet.ipynb",
         "xse_resnet34": "vision.models.xresnet.ipynb",
         "xse_resnet50": "vision.models.xresnet.ipynb",
         "xse_resnet101": "vision.models.xresnet.ipynb",
         "xse_resnext18": "vision.models.xresnet.ipynb",
         "xse_resnext34": "vision.models.xresnet.ipynb",
         "xse_resnext50": "vision.models.xresnet.ipynb",
         "xse_resnext101": "vision.models.xresnet.ipynb",
         "xeca_resnet18": "vision.models.xresnet.ipynb",
         "xeca_resnet34": "vision.models.xresnet.ipynb",
         "xeca_resnet50": "vision.models.xresnet.ipynb",
         "xeca_resnet101": "vision.models.xresnet.ipynb",
         "xeca_resnext18": "vision.models.xresnet.ipynb",
         "xeca_resnext34": "vision.models.xresnet.ipynb",
         "xeca_resnext50": "vision.models.xresnet.ipynb",
         "xeca_resnext101": "vision.models.xresnet.ipynb",
         "xsa_resnet18": "vision.models.xresnet.ipynb",
         "xsa_resnet34": "vision.models.xresnet.ipynb",
         "xsa_resnet50": "vision.models.xresnet.ipynb",
         "xsa_resnet101": "vision.models.xresnet.ipynb",
         "xsa_resnext18": "vision.models.xresnet.ipynb",
         "xsa_resnext34": "vision.models.xresnet.ipynb",
         "xsa_resnext50": "vision.models.xresnet.ipynb",
         "xsa_resnext101": "vision.models.xresnet.ipynb",
         "xta_resnet18": "vision.models.xresnet.ipynb",
         "xta_resnet34": "vision.models.xresnet.ipynb",
         "xta_resnet50": "vision.models.xresnet.ipynb",
         "xta_resnet101": "vision.models.xresnet.ipynb",
         "xta_resnext18": "vision.models.xresnet.ipynb",
         "xta_resnext34": "vision.models.xresnet.ipynb",
         "xta_resnext50": "vision.models.xresnet.ipynb",
         "xta_resnext101": "vision.models.xresnet.ipynb"}

modules = ["basics.py",
           "audio/core.py",
           "audio/data.py",
           "audio/augment.py",
           "audio/learner.py",
           "audio/mixup.py",
           "callback/casttotensor.py",
           "callback/channelslast.py",
           "callback/cutmixup.py",
           "callback/ema.py",
           "callback/lr_finder.py",
           "callback/progresize.py",
           "callback/simpleprofiler.py",
           "callback/tracker.py",
           "data/transforms.py",
           "losses.py",
           "metrics.py",
           "multiloss.py",
           "optimizers/adan.py",
           "optimizers/finetune.py",
           "schedulers/fit_flat_varied.py",
           "test_utils.py",
           "transform.py",
           "utils/__init__.py",
           "vision/augment/batch.py",
           "vision/augment/itemtensor.py",
           "vision/data.py",
           "vision/models/attention_modules.py",
           "vision/models/pooling.py",
           "vision/models/xresnet.py"]

doc_url = "https://fastxtend.benjaminwarner.dev/"

git_url = "https://github.com/warner-benjamin/fastxtend/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/audio.04_learner.ipynb.

# %% ../../nbs/audio.04_learner.ipynb 1
# Contains code from https://github.com/fastaudio/fastaudio:
# fastaudio - MIT License - Copyright (c) 2020 scart97

# %% ../../nbs/audio.04_learner.ipynb 3
from __future__ import annotations

from pathlib import Path

from fastcore.dispatch import retain_type

from fastai.callback.core import Callback
from fastai.callback.fp16 import MixedPrecision
from fastai.data.core import DataLoaders
from fastai.learner import Metric, Learner, defaults
from fastai.optimizer import Optimizer, OptimWrapper

from .core import TensorSpec, TensorMelSpec
from .data import MelSpectrogram, Spectrogram
from ..optimizer.fused import Adam
from ..imports import *

# %% auto 0
__all__ = ['StackSpecCallback', 'audio_learner']

# %% ../../nbs/audio.04_learner.ipynb 5
class StackSpecCallback(Callback):
    "Stacks tuples of TensorSpec or TensorMelSpec. ToDo: add resizing"
    order = MixedPrecision.order-1
    def before_batch(self):
        xb = L(self.xb)
        idx = xb.argwhere(lambda x: isinstance(x, (TensorSpec, TensorMelSpec)))
        ts = []
        for i in idx:
            ts.append(xb[i])
        stacked = torch.stack(ts, dim=2)
        xb = retain_type(torch.flatten(stacked, 1, 2), xb[i])
        self.learn.xb = tuple(xb)

# %% ../../nbs/audio.04_learner.ipynb 7
# TODO opt_func type hint should be Callable[..., Optimizer|OptimWrapper]|OptimWrapper
# but nbdev 2.3.9 currently doesn't support this type hint

def audio_learner(
    dls:DataLoaders, # `DataLoaders` containing fastai or PyTorch `DataLoader`s
    model:nn.Module, # PyTorch model for training or inference
    loss_func:nn.Module|Callable[..., Tensor]|None=None, # Loss function. Defaults to `dls` loss
    opt_func:Optimizer|OptimWrapper=Adam, # Optimization function for training
    lr:float|slice=defaults.lr, # Default learning rate
    splitter:Callable[[nn.Module], list[Tensor]]=trainable_params, # Split model into parameter groups. Defaults to one parameter group
    cbs:Listified[Callback]|None=None, # `Callback`s to add to `Learner`
    metrics:Listified[Metric]|None=None, # `Metric`s to calculate on validation set
    path:str|Path|None=None, # Parent directory to save, load, and export models. Defaults to `dls` `path`
    model_dir:str|Path='models', # Subdirectory to save and load models
    wd:float|None=None, # Default weight decay
    wd_bn_bias:bool=False, # Apply weight decay to normalization and bias parameters
    train_bn:bool=True, # Train frozen normalization layers
    moms:tuple[float,float,float]=(0.95,0.85,0.95), # Default momentum for schedulers
    default_cbs:bool=True # Include default `Callback`s
) -> Learner:
    "An Audio specific Learner that stacks tuples of `TensorSpec` or `TensorMelSpec`"
    detuple = False
    for i in range(len(dls.train.after_batch.fs)):
        if not detuple and isinstance(dls.train.after_batch[i], (Spectrogram, MelSpectrogram)):
            detuple = is_listy(dls.train.after_batch[i].n_fft)

    if detuple:
        if cbs is None:
            cbs = StackSpecCallback()
        else:
            cbs = L(cbs) + L(StackSpecCallback())

    return Learner(dls=dls, model=model, loss_func=loss_func, opt_func=opt_func, lr=lr, splitter=splitter, cbs=cbs,
                    metrics=metrics, path=path, model_dir=model_dir, wd=wd, wd_bn_bias=wd_bn_bias, train_bn=train_bn,
                    moms=moms, default_cbs=default_cbs)
